import pandas as pd
import numpy as np
from fair_perf_ml import data_bias, model_bias, model_perf
from numpy.typing import NDArray
from typing import Tuple
import argparse


def generate_binary_data(len: int) -> Tuple[NDArray, NDArray, NDArray]:
    np.random.seed(1)
    proba = np.random.rand(len)
    pred = np.where(proba >= 0.5, 1.0, 0.0)
    np.random.seed(57)
    true = np.where(np.random.rand(len) >= 0.5, 1.0, 0.0)
    df = pd.concat(
        [
            pd.Series(data=true, name="true"),
            pd.Series(data=pred, name="pred"),
            pd.Series(data=proba, name="proba"),
        ],
        axis=1,
    )
    df.to_csv("perf_data.csv", header=True, index=False)
    return true, pred, proba


def generate_synthetic_scores(row: int) -> float:
    op = 1 if np.random.random() > 0.5 else -1
    return row + (2 * np.random.random() * op)


def generate_test_set(df: pd.DataFrame) -> pd.DataFrame:
    df["preds"] = df.rings.apply(generate_synthetic_scores)
    df.to_csv("test_data.csv", header=True, index=False)
    return df


def get_data() -> pd.DataFrame:
    headers = [
        "sex",
        "length",
        "diameter",
        "height",
        "whole_weight",
        "shucked_weight",
        "viscera_weight",
        "shell_weight",
        "rings",
    ]
    """
    Using abalone dataset to test
    """
    return pd.read_csv("abalone.data", names=headers)


def test_db_numpy(bl_df, runtime_test) -> bool:
    """
    testing using a numpy array with the db methods
    """
    res = [False] * 3
    db_bl = data_bias.perform_analysis(
        bl_df["sex"].to_numpy(), bl_df["rings"].to_numpy(), "M", 15
    )

    res[0] = db_bl == {
        "ClassImbalance": 0.2545889914035797,
        "DifferenceInProportionOfLabels": 0.039049260318279266,
        "KlDivergence": 0.007977087050676346,
        "JsDivergence": 0.002021726220846176,
        "LpNorm": 0.23192767798900604,
        "TotalVarationDistance": 0.039049260318279266,
        "KolmorogvSmirnov": 0.039049260318279266,
    }

    db_runtime = data_bias.perform_analysis(
        runtime_test["sex"].to_numpy(), runtime_test["rings"].to_numpy(), "M", 15
    )
    res[1] = db_runtime == {
        "ClassImbalance": 0.28862276673316956,
        "DifferenceInProportionOfLabels": 0.026097401976585388,
        "KlDivergence": 0.006829630583524704,
        "JsDivergence": 0.0016257409006357193,
        "LpNorm": 0.11823293566703796,
        "TotalVarationDistance": 0.026097401976585388,
        "KolmorogvSmirnov": 0.026097401976585388,
    }

    runtime_check = data_bias.runtime_comparison(db_bl, db_runtime, 0.15)

    res[2] = runtime_check == {"passed": True}

    return all(res)


def test_db_numpy_partial(bl_df, runtime_test) -> bool:
    """
    testing using a numpy array with the db methods
    """
    res = [False] * 3
    db_bl = data_bias.perform_analysis(
        bl_df["sex"].to_numpy(), bl_df["rings"].to_numpy(), "M", 15
    )

    res[0] = db_bl == {
        "ClassImbalance": 0.2545889914035797,
        "DifferenceInProportionOfLabels": 0.039049260318279266,
        "KlDivergence": 0.007977087050676346,
        "JsDivergence": 0.002021726220846176,
        "LpNorm": 0.23192767798900604,
        "TotalVarationDistance": 0.039049260318279266,
        "KolmorogvSmirnov": 0.039049260318279266,
    }

    db_runtime = data_bias.perform_analysis(
        runtime_test["sex"].to_numpy(), runtime_test["rings"].to_numpy(), "M", 15
    )

    res[1] = db_runtime == {
        "ClassImbalance": 0.28862276673316956,
        "DifferenceInProportionOfLabels": 0.026097401976585388,
        "KlDivergence": 0.006829630583524704,
        "JsDivergence": 0.0016257409006357193,
        "LpNorm": 0.11823293566703796,
        "TotalVarationDistance": 0.026097401976585388,
        "KolmorogvSmirnov": 0.026097401976585388,
    }

    runtime_check = data_bias.partial_runtime_comparison(
        db_bl,
        db_runtime,
        [
            "ClassImbalance",
            "DifferenceInProportionOfLabels",
            "KlDivergence",
            "JsDivergence",
        ],
        0.15,
    )

    res[2] = runtime_check == {"passed": True}

    return all(res)


def test_db_list(bl_df, runtime_test) -> bool:
    """
    testing using a numpy array with the db methods
    """
    res = [False] * 3
    db_bl = data_bias.perform_analysis(
        bl_df["sex"].to_numpy(), bl_df["rings"].to_numpy(), "M", 15
    )

    res[0] = db_bl == {
        "ClassImbalance": 0.2545889914035797,
        "DifferenceInProportionOfLabels": 0.039049260318279266,
        "KlDivergence": 0.007977087050676346,
        "JsDivergence": 0.002021726220846176,
        "LpNorm": 0.23192767798900604,
        "TotalVarationDistance": 0.039049260318279266,
        "KolmorogvSmirnov": 0.039049260318279266,
    }

    db_runtime = data_bias.perform_analysis(
        runtime_test["sex"].to_numpy(), runtime_test["rings"].to_numpy(), "M", 15
    )
    res[1] = db_runtime == {
        "ClassImbalance": 0.28862276673316956,
        "DifferenceInProportionOfLabels": 0.026097401976585388,
        "KlDivergence": 0.006829630583524704,
        "JsDivergence": 0.0016257409006357193,
        "LpNorm": 0.11823293566703796,
        "TotalVarationDistance": 0.026097401976585388,
        "KolmorogvSmirnov": 0.026097401976585388,
    }

    runtime_check = data_bias.runtime_comparison(db_bl, db_runtime, 0.15)

    res[2] = runtime_check == {"passed": True}

    return all(res)


def test_mb_numpy(bl_df, runtime_test) -> bool:
    res = [False] * 3

    bl = model_bias.perform_analysis(
        bl_df["sex"].to_numpy(),
        bl_df["rings"].to_numpy(),
        bl_df["preds"].to_numpy(),
        "M",
        15,
        15.0,
    )

    res[0] = bl == {
        "DifferenceInPositivePredictedLabels": 0.002093970775604248,
        "DisparateImpact": 0.8409091234207153,
        "AccuracyDifference": -8.034706115722656e-05,
        "RecallDifference": 0.04676508903503418,
        "DifferenceInConditionalAcceptance": -0.0026617050170898438,
        "DifferenceInAcceptanceRate": 0.05057328939437866,
        "SpecialityDifference": -0.0018883943557739258,
        "DifferenceInConditionalRejection": 0.005205392837524414,
        "DifferenceInRejectionRate": 0.0032941699028015137,
        "TreatmentEquity": -0.8666665554046631,
        "ConditionalDemographicDesparityPredictedLabels": 0.09311360120773315,
        "GeneralizedEntropy": 139933.03125,
    }

    runtime = model_bias.perform_analysis(
        runtime_test["sex"].to_numpy(),
        runtime_test["rings"].to_numpy(),
        runtime_test["preds"].to_numpy(),
        "M",
        15,
        15.0,
    )

    res[1] = runtime == {
        "DifferenceInPositivePredictedLabels": -0.09821432828903198,
        "DisparateImpact": 0.7872340083122253,
        "AccuracyDifference": -0.00829547643661499,
        "RecallDifference": -0.0654761791229248,
        "DifferenceInConditionalAcceptance": 0.11385858058929443,
        "DifferenceInAcceptanceRate": 0.025301873683929443,
        "SpecialityDifference": 0.0005710124969482422,
        "DifferenceInConditionalRejection": 0.008004844188690186,
        "DifferenceInRejectionRate": 0.008531749248504639,
        "TreatmentEquity": -1.4666666984558105,
        "ConditionalDemographicDesparityPredictedLabels": 0.0892782211303711,
        "GeneralizedEntropy": 27914.095703125,
    }

    runtime_check = model_bias.runtime_comparison(bl, runtime, 0.15)

    res[2] = runtime_check == {
        "passed": False,
        "failReport": {
            "DifferenceInPositivePredictedLabels": "Exceed baseline by: 0.09612036",
            "RecallDifference": "Exceed baseline by: 0.01871109",
            "AccuracyDifference": "Exceed baseline by: 0.008215129",
            "DifferenceInConditionalRejection": "Exceed baseline by: 0.0027994514",
            "DifferenceInConditionalAcceptance": "Exceed baseline by: 0.111196876",
            "DifferenceInRejectionRate": "Exceed baseline by: 0.0052375793",
            "TreatmentEquity": "Exceed baseline by: 0.60000014",
        },
    }
    return all(res)


def test_mb_numpy_partial(bl_df, runtime_test) -> bool:
    res = [False] * 3
    bl = model_bias.perform_analysis(
        bl_df["sex"].to_numpy(),
        bl_df["rings"].to_numpy(),
        bl_df["preds"].to_numpy(),
        "M",
        15,
        15.0,
    )

    res[0] = bl == {
        "DifferenceInPositivePredictedLabels": 0.002093970775604248,
        "DisparateImpact": 0.8409091234207153,
        "AccuracyDifference": -8.034706115722656e-05,
        "RecallDifference": 0.04676508903503418,
        "DifferenceInConditionalAcceptance": -0.0026617050170898438,
        "DifferenceInAcceptanceRate": 0.05057328939437866,
        "SpecialityDifference": -0.0018883943557739258,
        "DifferenceInConditionalRejection": 0.005205392837524414,
        "DifferenceInRejectionRate": 0.0032941699028015137,
        "TreatmentEquity": -0.8666665554046631,
        "ConditionalDemographicDesparityPredictedLabels": 0.09311360120773315,
        "GeneralizedEntropy": 139933.03125,
    }

    runtime = model_bias.perform_analysis(
        runtime_test["sex"].to_numpy(),
        runtime_test["rings"].to_numpy(),
        runtime_test["preds"].to_numpy(),
        "M",
        15,
        15.0,
    )

    res[1] = runtime == {
        "DifferenceInPositivePredictedLabels": -0.09821432828903198,
        "DisparateImpact": 0.7872340083122253,
        "AccuracyDifference": -0.00829547643661499,
        "RecallDifference": -0.0654761791229248,
        "DifferenceInConditionalAcceptance": 0.11385858058929443,
        "DifferenceInAcceptanceRate": 0.025301873683929443,
        "SpecialityDifference": 0.0005710124969482422,
        "DifferenceInConditionalRejection": 0.008004844188690186,
        "DifferenceInRejectionRate": 0.008531749248504639,
        "TreatmentEquity": -1.4666666984558105,
        "ConditionalDemographicDesparityPredictedLabels": 0.0892782211303711,
        "GeneralizedEntropy": 27914.095703125,
    }

    runtime_check = model_bias.partial_runtime_comparison(
        bl,
        runtime,
        [
            "DifferenceInPositivePredictedLabels",
            "DisparateImpact",
            "AccuracyDifference",
            "RecallDifference",
            "DifferenceInConditionalAcceptance",
        ],
        0.15,
    )

    res[2] = runtime_check == {
        "passed": False,
        "failReport": {
            "DifferenceInPositivePredictedLabels": "Exceed baseline by: 0.09612036",
            "AccuracyDifference": "Exceed baseline by: 0.008215129",
            "DifferenceInConditionalAcceptance": "Exceed baseline by: 0.111196876",
            "RecallDifference": "Exceed baseline by: 0.01871109",
        },
    }
    return all(res)


def test_perf_reg_numpy(y_pred, y_true) -> bool:
    res = [False] * 4
    l = int(y_pred.size * 0.7)
    bl_true = y_true[:l]
    rt_true = y_true[l:]
    bl_pred = y_pred[:l]
    rt_pred = y_pred[l:]

    bl = model_perf.linear_regression_analysis(y_true=bl_true, y_pred=bl_pred)

    res[0] = bl == {
        "modelType": "LinearRegression",
        "performanceData": {
            "RootMeanSquaredError": 1.1617724895477295,
            "MeanSquaredError": 1.349715232849121,
            "MeanAbsoluteError": 1.0100986957550049,
            "RSquared": 0.1202625185251236,
            "MaxError": 1.999868392944336,
            "MeanSquaredLogError": 204887.609375,
            "RootMeanSquaredLogError": 24472.158203125,
            "MeanAbsolutePercentageError": 11.345174789428711,
        },
    }
    runtime = model_perf.linear_regression_analysis(y_true=rt_true, y_pred=rt_pred)

    res[1] = runtime == {
        "modelType": "LinearRegression",
        "performanceData": {
            "RootMeanSquaredError": 1.133068323135376,
            "MeanSquaredError": 1.283843755722046,
            "MeanAbsoluteError": 0.9810680150985718,
            "RSquared": 0.15180453658103943,
            "MaxError": 1.999338150024414,
            "MeanSquaredLogError": 54827.40234375,
            "RootMeanSquaredLogError": 8291.77734375,
            "MeanAbsolutePercentageError": 10.8325834274292,
        },
    }
    partial = model_perf.partial_runtime_check(
        baseline=bl,
        latest=runtime,
        metrics=[
            "RootMeanSquaredError",
            "MeanSquaredError",
            "MeanAbsoluteError",
            "RSquared",
        ],
    )


    res[2] = partial == {
        "passed": False,
        "failReport": {"RSquared": "Exceeded threshold by 0.031542018"},
    }
    full_res = model_perf.runtime_check_full(baseline=bl, latest=runtime)
    res[3] = full_res == {
        "passed": False,
        "failReport": {"RSquared": "Exceeded threshold by 0.031542018"},
    }
    return all(res)


def test_perf_reg_list(y_pred, y_true):
    res = [False] * 3
    l = int(len(y_pred) * 0.7)
    bl_true = y_true[:l]
    rt_true = y_true[l:]
    bl_pred = y_pred[:l]
    rt_pred = y_pred[l:]

    bl = model_perf.linear_regression_analysis(y_true=bl_true, y_pred=bl_pred)

    res[0] = bl == {
        "modelType": "LinearRegression",
        "performanceData": {
            "RootMeanSquaredError": 1.1617724895477295,
            "MeanSquaredError": 1.349715232849121,
            "MeanAbsoluteError": 1.0100986957550049,
            "RSquared": 0.1202625185251236,
            "MaxError": 1.999868392944336,
            "MeanSquaredLogError": 204887.609375,
            "RootMeanSquaredLogError": 24472.158203125,
            "MeanAbsolutePercentageError": 11.345174789428711,
        },
    }

    runtime = model_perf.linear_regression_analysis(y_true=rt_true, y_pred=rt_pred)

    res[1] = runtime == {
        "modelType": "LinearRegression",
        "performanceData": {
            "RootMeanSquaredError": 1.133068323135376,
            "MeanSquaredError": 1.283843755722046,
            "MeanAbsoluteError": 0.9810680150985718,
            "RSquared": 0.15180453658103943,
            "MaxError": 1.999338150024414,
            "MeanSquaredLogError": 54827.40234375,
            "RootMeanSquaredLogError": 8291.77734375,
            "MeanAbsolutePercentageError": 10.8325834274292,
        },
    }

    full_res = model_perf.runtime_check_full(baseline=bl, latest=runtime)
    res[2] = full_res == {
        "passed": False,
        "failReport": {"RSquared": "Exceeded threshold by 0.031542018"},
    }

    return all(res)

def test_perf_reg_classification_numpy(y_pred, y_true) -> bool:
    res = [False] * 3
    l = int(y_pred.size * 0.7)
    bl_true = y_true[:l]
    pred_true = y_true[l:]
    bl_pred = y_pred[:l]
    pred_pred = y_pred[l:]

    bl = model_perf.binary_classification_analysis(y_true=bl_true, y_pred=bl_pred)
    res[0] = bl == {
        "modelType": "BinaryClassification",
        "performanceData": {
            "BalancedAccuracy": 0.12866388261318207,
            "PrecisionPositive": 0.5059422850608826,
            "PrecisionNegative": 0.5086554884910583,
            "RecallPositive": 0.5121741890907288,
            "RecallNegative": 0.5024223327636719,
            "Accuracy": 0.5072857141494751,
            "F1Score": 0.5090391635894775,
        },
    }

    runtime = model_perf.binary_classification_analysis(
        y_true=pred_true, y_pred=pred_pred
    )


    res[1] = runtime == {
        "modelType": "BinaryClassification",
        "performanceData": {
            "BalancedAccuracy": 0.11985205113887787,
            "PrecisionPositive": 0.48915988206863403,
            "PrecisionNegative": 0.49015748500823975,
            "RecallPositive": 0.48165443539619446,
            "RecallNegative": 0.49766820669174194,
            "Accuracy": 0.4896666705608368,
            "F1Score": 0.4853781461715698,
        },
    }

    eval_res = model_perf.runtime_check_full(baseline=bl, latest=runtime)

    res[2] = eval_res == {"passed": True}

    return all(res)


def test_perf_reg_classification_list(y_pred, y_true) -> bool:
    res = [False] * 3
    l = int(len(y_pred) * 0.7)
    bl_true = y_true[:l]
    pred_true = y_true[l:]
    bl_pred = y_pred[:l]
    pred_pred = y_pred[l:]

    bl = model_perf.binary_classification_analysis(y_true=bl_true, y_pred=bl_pred)
    res[0] = bl == {
        "modelType": "BinaryClassification",
        "performanceData": {
            "BalancedAccuracy": 0.12866388261318207,
            "PrecisionPositive": 0.5059422850608826,
            "PrecisionNegative": 0.5086554884910583,
            "RecallPositive": 0.5121741890907288,
            "RecallNegative": 0.5024223327636719,
            "Accuracy": 0.5072857141494751,
            "F1Score": 0.5090391635894775,
        },
    }

    runtime = model_perf.binary_classification_analysis(
        y_true=pred_true, y_pred=pred_pred
    )


    res[1] = runtime == {
        "modelType": "BinaryClassification",
        "performanceData": {
            "BalancedAccuracy": 0.11985205113887787,
            "PrecisionPositive": 0.48915988206863403,
            "PrecisionNegative": 0.49015748500823975,
            "RecallPositive": 0.48165443539619446,
            "RecallNegative": 0.49766820669174194,
            "Accuracy": 0.4896666705608368,
            "F1Score": 0.4853781461715698,
        },
    }

    eval_res = model_perf.runtime_check_full(baseline=bl, latest=runtime)

    res[2] = eval_res == {"passed": True}

    return all(res)


def test_perf_logisitc_reg_numpy(y_pred, y_true) -> bool:
    res = [False] * 3
    l = int(y_pred.size * 0.7)
    bl_true = y_true[:l]
    rt_true = y_true[l:]
    bl_pred = y_pred[:l]
    rt_pred = y_pred[l:]

    bl = model_perf.logistic_regression_analysis(y_true=bl_true, y_pred=bl_pred)
    res[0] = bl == {
        "modelType": "LogisticRegression",
        "performanceData": {
            "BalancedAccuracy": 0.12866388261318207,
            "PrecisionPositive": 0.5059422850608826,
            "PrecisionNegative": 0.5086554884910583,
            "RecallPositive": 0.5121741890907288,
            "RecallNegative": 0.5024223327636719,
            "Accuracy": 0.5072857141494751,
            "F1Score": 0.5090391635894775,
            "LogLoss": 0.4283928871154785,
        },
    }

    runtime = model_perf.logistic_regression_analysis(y_true=rt_true, y_pred=rt_pred)

    res[1] = runtime == {
        "modelType": "LogisticRegression",
        "performanceData": {
            "BalancedAccuracy": 0.11985205113887787,
            "PrecisionPositive": 0.48915988206863403,
            "PrecisionNegative": 0.49015748500823975,
            "RecallPositive": 0.48165443539619446,
            "RecallNegative": 0.49766820669174194,
            "Accuracy": 0.4896666705608368,
            "F1Score": 0.4853781461715698,
            "LogLoss": 0.4418107867240906,
        },
    }

    eval_res = model_perf.runtime_check_full(baseline=bl, latest=runtime)

    res[2] = eval_res == {"passed": True}
    return all(res)


def test_perf_logisitc_reg_list(y_pred, y_true):
    res = [False] * 3
    l = int(len(y_pred) * 0.7)
    bl_true = y_true[:l]
    rt_true = y_true[l:]
    bl_pred = y_pred[:l]
    rt_pred = y_pred[l:]

    bl = model_perf.logistic_regression_analysis(y_true=bl_true, y_pred=bl_pred)
    res[0] = bl == {
        "modelType": "LogisticRegression",
        "performanceData": {
            "BalancedAccuracy": 0.12866388261318207,
            "PrecisionPositive": 0.5059422850608826,
            "PrecisionNegative": 0.5086554884910583,
            "RecallPositive": 0.5121741890907288,
            "RecallNegative": 0.5024223327636719,
            "Accuracy": 0.5072857141494751,
            "F1Score": 0.5090391635894775,
            "LogLoss": 0.4283928871154785,
        },
    }

    runtime = model_perf.logistic_regression_analysis(y_true=rt_true, y_pred=rt_pred)

    res[1] = runtime == {
        "modelType": "LogisticRegression",
        "performanceData": {
            "BalancedAccuracy": 0.11985205113887787,
            "PrecisionPositive": 0.48915988206863403,
            "PrecisionNegative": 0.49015748500823975,
            "RecallPositive": 0.48165443539619446,
            "RecallNegative": 0.49766820669174194,
            "Accuracy": 0.4896666705608368,
            "F1Score": 0.4853781461715698,
            "LogLoss": 0.4418107867240906,
        },
    }

    eval_res = model_perf.runtime_check_full(baseline=bl, latest=runtime)

    res[2] = eval_res == {"passed": True}
    return all(res)


def test_mb_list(bl_df, runtime_test):
    res = [False] * 3

    bl = model_bias.perform_analysis(
        bl_df["sex"].to_list(),
        bl_df["rings"].to_list(),
        bl_df["preds"].to_list(),
        "M",
        15,
        15.0,
    )


    res[0] = bl == {
        "DifferenceInPositivePredictedLabels": 0.002093970775604248,
        "DisparateImpact": 0.8409091234207153,
        "AccuracyDifference": -8.034706115722656e-05,
        "RecallDifference": 0.04676508903503418,
        "DifferenceInConditionalAcceptance": -0.0026617050170898438,
        "DifferenceInAcceptanceRate": 0.05057328939437866,
        "SpecialityDifference": -0.0018883943557739258,
        "DifferenceInConditionalRejection": 0.005205392837524414,
        "DifferenceInRejectionRate": 0.0032941699028015137,
        "TreatmentEquity": -0.8666665554046631,
        "ConditionalDemographicDesparityPredictedLabels": 0.09311360120773315,
        "GeneralizedEntropy": 139933.03125,
    }
    runtime = model_bias.perform_analysis(
        runtime_test["sex"].to_list(),
        runtime_test["rings"].to_list(),
        runtime_test["preds"].to_list(),
        "M",
        15,
        15.0,
    )


    res[1] = runtime == {
        "DifferenceInPositivePredictedLabels": -0.09821432828903198,
        "DisparateImpact": 0.7872340083122253,
        "AccuracyDifference": -0.00829547643661499,
        "RecallDifference": -0.0654761791229248,
        "DifferenceInConditionalAcceptance": 0.11385858058929443,
        "DifferenceInAcceptanceRate": 0.025301873683929443,
        "SpecialityDifference": 0.0005710124969482422,
        "DifferenceInConditionalRejection": 0.008004844188690186,
        "DifferenceInRejectionRate": 0.008531749248504639,
        "TreatmentEquity": -1.4666666984558105,
        "ConditionalDemographicDesparityPredictedLabels": 0.0892782211303711,
        "GeneralizedEntropy": 27914.095703125,
    }
    runtime_check = model_bias.runtime_comparison(bl, runtime, 0.15)

    res[2] = runtime_check == {
        "passed": False,
        "failReport": {
            "DifferenceInPositivePredictedLabels": "Exceed baseline by: 0.09612036",
            "AccuracyDifference": "Exceed baseline by: 0.008215129",
            "DifferenceInConditionalRejection": "Exceed baseline by: 0.0027994514",
            "RecallDifference": "Exceed baseline by: 0.01871109",
            "TreatmentEquity": "Exceed baseline by: 0.60000014",
            "DifferenceInConditionalAcceptance": "Exceed baseline by: 0.111196876",
            "DifferenceInRejectionRate": "Exceed baseline by: 0.0052375793",
        },
    }

    return all(res)


def eval_str_to_bool(v: str) -> bool:
    if v.lower() not in ["true", "false"]:
        raise ValueError("Invalid value")
    return v.lower() == "true"


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--generate-data", type=eval_str_to_bool, default=False)
    args = parser.parse_args()
    generate = args.generate_data
    if generate:
        df = get_data()
        df = generate_test_set(df)
        bin_true, bin_pred, bin_proba = generate_binary_data(10000)
    else:
        df = pd.read_csv("test_data.csv")
        perf_data = pd.read_csv("perf_data.csv")
        bin_true, bin_pred, bin_proba = (
            perf_data["true"].to_numpy(),
            perf_data["pred"].to_numpy(),
            perf_data["proba"].to_numpy(),
        )

    reg_true = df["rings"].to_numpy().copy()
    reg_pred = df["preds"].to_numpy().copy()

    bl_df = df.iloc[: int(0.60 * df.shape[0]), :]
    runtime_test = df.iloc[int(0.60 * df.shape[0]) + 1 :, :]
    print("TESTING DATA BIAS WITH NUMPY ARRAYS...")
    assert test_db_numpy(bl_df, runtime_test)
    print("passed...")

    print("TESTING DATA BIAS WITH NUMPY ARRAYS partial...")
    assert test_db_numpy_partial(bl_df, runtime_test)
    print("passed...")

    print("TESTING DATA BIAS WITH LIST ARRAYS...")
    assert test_db_list(bl_df, runtime_test)
    print("passed...")

    print("TESTING MB with numpy...")
    assert test_mb_numpy(bl_df, runtime_test)
    print("passed...")

    print("TESTING MB with numpy PARTIAL...")
    assert test_mb_numpy_partial(bl_df, runtime_test)
    print("passed...")

    print("TESTING MB with lists...")
    assert test_mb_list(bl_df, runtime_test)
    print("passed...")

    print("TESTING PERF WITH NUMPY ARRAYS")
    assert test_perf_reg_numpy(reg_pred, reg_true)
    print("passed...")

    print("TESTING PERF WITH LISTS")
    assert test_perf_reg_list(reg_pred.tolist(), reg_true.tolist())
    print("passed...")

    print("TESTING Classification PERF WITH NUMPY")
    assert test_perf_reg_classification_numpy(bin_pred, bin_true)
    print("passed...")

    print("TESTING Classification PERF WITH LISTS")
    assert test_perf_reg_classification_list(bin_pred.tolist(), bin_true.tolist())
    print("passed...")

    print("TESTING logisitc PERF WITH numpy")
    assert test_perf_logisitc_reg_numpy(bin_proba, bin_true)
    print("passed...")

    print("TESTING logisitc PERF WITH list")
    assert test_perf_logisitc_reg_list(bin_proba.tolist(), bin_true.tolist())
    print("passed...")



from ._fair_perf_ml import (
    model_performance_regression,
    model_performance_classification,
    model_performance_logisitic_regression,
    model_performance_runtime_entry_full,
    model_performance_runtime_entry_partial,
)
from .models import (
    ModelType,
    ModelPerformance,
    LinearRegressionReport,
    LogisticRegressionReport,
    BinaryClassificationReport,
)
from ._internal import check_and_convert_type
from numpy.typing import NDArray
from typing import Union, List, Optional
import orjson


class DifferentModelTypes(Exception):
    pass


class InvalidMetricsBody(Exception):
    pass


def linear_regression_analysis(
    y_true: Union[NDArray, List[Union[int, float]]],  # pyright: ignore
    y_pred: Union[NDArray, List[Union[int, float]]],  # pyright: ignore
) -> dict:
    y_true: NDArray = check_and_convert_type(y_true)  # pyright: ignore
    y_pred: NDArray = check_and_convert_type(y_pred)  # pyright: ignore
    res: dict = model_performance_regression(y_true=y_true, y_pred=y_pred)
    return ModelPerformance(
        modelType=ModelType.LinearRegression,
        performanceData=LinearRegressionReport(**res),
    ).model_dump()


def logistic_regression_analysis(
    y_true: Union[NDArray, List[Union[int, float]]],  # pyright: ignore
    y_pred: Union[NDArray, List[Union[int, float]]],  # pyright: ignore
    decision_threshold: Optional[float] = 0.5,
) -> dict:
    y_true: NDArray = check_and_convert_type(y_true)  # pyright: ignore
    y_pred: NDArray = check_and_convert_type(y_pred)  # pyright: ignore
    res: dict = model_performance_logisitic_regression(
        y_true=y_true, y_pred=y_pred, decision_threshold=decision_threshold
    )
    return ModelPerformance(
        modelType=ModelType.LogisticRegression,
        performanceData=LogisticRegressionReport(**res),
    ).model_dump()


def binary_classification_analysis(
    y_true: Union[NDArray, List[Union[int, float]]],  # pyright: ignore
    y_pred: Union[NDArray, List[Union[int, float]]],  # pyright: ignore
) -> dict:
    y_true: NDArray = check_and_convert_type(y_true)  # pyright: ignore
    y_pred: NDArray = check_and_convert_type(y_pred)  # pyright: ignore
    res: dict = model_performance_classification(y_true=y_true, y_pred=y_pred)
    return ModelPerformance(
        modelType=ModelType.BinaryClassification,
        performanceData=BinaryClassificationReport(**res),
    ).model_dump()


def runtime_check_full(
    latest: dict, baseline: dict, threshold: Optional[float] = 0.10
) -> dict:
    model_type = baseline.get("modelType")
    if model_type != latest.get("modelType"):
        raise DifferentModelTypes("Models types do not match")
    latest_perf = latest.get("performanceData")
    baseline_perf = baseline.get("performanceData")
    if any([model_type is None, latest_perf is None, baseline_perf is None]):
        raise InvalidMetricsBody("Invalid metrics body")
    perf = model_performance_runtime_entry_full(
        model_type=model_type,
        latest=latest_perf,
        baseline=baseline_perf,
        threshold=threshold,
    )
    return orjson.loads(perf)


def partial_runtime_check(
    latest: dict, baseline: dict, metrics: List[str], threshold: Optional[float] = 0.10
) -> dict:
    model_type = baseline.get("modelType")
    latest_perf = latest.get("performanceData")
    baseline_perf = baseline.get("performanceData")
    if any([model_type is None, latest_perf is None, baseline_perf is None]):
        raise InvalidMetricsBody("Invalid metrics body")
    perf = model_performance_runtime_entry_partial(
        model_type=model_type,
        latest=latest_perf,
        baseline=baseline_perf,
        evaluation_metrics=metrics,
        threshold=threshold,
    )
    return orjson.loads(perf)



from ._fair_perf_ml import (
    model_bias_analyzer,
    model_bias_runtime_check,
    model_bias_partial_check,
)
from ._internal import check_and_convert_type
from .models import ModelBiasBaseline
from numpy.typing import NDArray
from typing import List, Union, Optional
import orjson


def perform_analysis(
    feature: Union[List[Union[str, float, int]], NDArray],  # pyright: ignore
    ground_truth: Union[List[Union[str, float, int]], NDArray],  # pyright: ignore
    predictions: Union[List[Union[str, float, int]], NDArray],  # pyright: ignore
    feature_label_or_threshold: Union[str, float, int],
    ground_truth_label_or_threshold: Union[str, float, int],
    prediction_label_or_threshold: Union[str, float, int],
) -> dict[str, float]:
    """
    interface into rust class
    makes sure we are passing numpy arrays to the rust function
    Args:
        feature: Union[List[Union[str, float, int]], NDArray] -> the feature data
            most efficient to pass as numpy array
        ground_truth: Union[List[Union[str, float, int]], NDArray] -> the ground truth data
            most efficient to pass as numpy array
        predictions: Union[List[Union[str, float, int]], NDArray] -> the prediction data
            most efficient to pass as numpy array
        feature_label_or_threshold: Union[str, float, int] -> segmentation parameter for the feature
        ground_truth_label_or_threshold: Union[str, float, int] -> segmenation parameter for ground truth
        prediction_label_or_threshold: Union[str, float, int] -> segmenation parameter for predictions
    """
    feature: NDArray = check_and_convert_type(feature)
    ground_truth: NDArray = check_and_convert_type(ground_truth)
    predictions: NDArray = check_and_convert_type(predictions)

    res: dict[str, float] = model_bias_analyzer(
        feature_array=feature,
        ground_truth_array=ground_truth,
        prediction_array=predictions,
        feature_label_or_threshold=feature_label_or_threshold,
        ground_truth_label_or_threshold=ground_truth_label_or_threshold,
        prediction_label_or_threshold=prediction_label_or_threshold,
    )

    # for nice formatting
    return ModelBiasBaseline(**res).model_dump()


def runtime_comparison(
    baseline: dict, comparison: dict, threshold: Optional[float] = None
) -> dict[str, str]:
    """
    interface into rust module
    serves to nicely formats the return as dicts are ordered and hashmaps are not
    Args:
        baseline: dict -> the result from calling perform_analysis on the baseline data
        latest: dict -> the current data for comparison from calling perform_analysis
        threshold: Optionl[float]=None -> the comparison threshold, defaults to 0.10 in rust mod
    Returns:
        dict
    """
    res: str = model_bias_runtime_check(
        baseline=baseline, latest=comparison, threshold=threshold
    )

    # for nice formatting
    return orjson.loads(res)


def partial_runtime_comparison(
    baseline: dict,
    comparison: dict,
    metrics: List[str],
    threshold: Optional[float] = None,
) -> dict[str, str]:
    """
    interface into rust module
    data body validation will happen within the rust logic
    serves to nicely formats the return as dicts are ordered and hashmaps are not
    Args:
        baseline: dict -> the result from calling perform_analysis on the baseline data
        latest: dict -> the current data for comparison from calling perform_analysis
        metrics: List[str] -> the list of metrics we want to evaluate on
        threshold: Optionl[float]=None -> the comparison threshold, defaults to 0.10 in rust mod
    Returns:
        dict
    """
    res: str = model_bias_partial_check(
        baseline=baseline, latest=comparison, metrics=metrics, threshold=threshold
    )

    # for nice formatting
    return orjson.loads(res)



import numpy as np
from numpy.typing import NDArray
from typing import List, Union


def check_and_convert_type(
    arr: Union[List[Union[str, float, int]], NDArray]
) -> NDArray:
    if _is_numpy(arr):
        return arr  # pyright: ignore
    if not _is_uniform_type(arr):  # pyright: ignore
        raise ValueError(f"Array needs to be of uniform type when of type {list}")
    return _convert_obj_type(arr)  # pyright: ignore


def _is_numpy(arr: Union[List[Union[str, float, int]], NDArray]) -> bool:
    return isinstance(arr, np.ndarray)


def _is_uniform_type(arr: List[Union[str, float, int]]) -> bool:
    if not isinstance(arr, list):
        return False
    T = type(arr[0])
    return all([True if type(item) == T else False for item in arr])


def _convert_obj_type(arr: List[Union[str, float, int]]) -> NDArray:
    return np.array(arr)



from ._fair_perf_ml import (
    data_bias_analyzer,
    data_bias_runtime_check,
    data_bias_partial_check,
)
from typing import Union, Optional, List
from numpy.typing import NDArray
from orjson import loads
from .models import DataBiasBaseline
from ._internal import check_and_convert_type


def perform_analysis(
    feature: Union[List[Union[str, float, int]], NDArray],  # pyright: ignore
    ground_truth: Union[List[Union[str, float, int]], NDArray],  # pyright: ignore
    feature_label_or_threshold: Union[str, float, int],
    ground_truth_label_or_threshold: Union[str, float, int],
) -> dict[str, float]:
    """
    interface into rust class
    makes sure we are passing numpy arrays to the rust function
    Args:
        feature: Union[List[Union[str, float, int]], NDArray] -> the feature data
            most efficient to pass as numpy array
        ground_truth: Union[List[Union[str, float, int]], NDArray] -> the ground truth data
            most efficient to pass as numpy array
        feature_label_or_threshold: Union[str, float, int] -> segmentation parameter for the feature
        ground_truth_label_or_threshold: Union[str, float, int] -> segmenation parameter for ground truth
    """
    # want to pass numpy arrays to rust
    # type resolution in rust mod depends on numpy arrays
    # ignoring lsp message because we are deliberately changing the type on these values
    feature: NDArray = check_and_convert_type(feature)
    ground_truth: NDArray = check_and_convert_type(ground_truth)

    res: dict[str, float] = data_bias_analyzer(
        feature_array=feature,
        ground_truth_array=ground_truth,
        feature_label_or_threshold=feature_label_or_threshold,
        ground_truth_label_or_threshold=ground_truth_label_or_threshold,
    )

    # simply for nice formatting
    return DataBiasBaseline(**res).model_dump()


def runtime_comparison(
    baseline: dict[str, float],
    latest: dict[str, float],
    threshold: Optional[float] = None,
) -> dict[str, str]:
    """
    interface into rust module
    serves to nicely formats the return as dicts are ordered and hashmaps are not
    Args:
        baseline: dict -> the result from calling perform_analysis on the baseline data
        latest: dict -> the current data for comparison from calling perform_analysis
        threshold: Optionl[float]=None -> the comparison threshold, defaults to 0.10 in rust mod
    Returns:
        dict
    """
    res: str = (
        data_bias_runtime_check(baseline=baseline, latest=latest, threshold=threshold)
        if threshold
        else data_bias_runtime_check(baseline=baseline, latest=latest)
    )
    # for nicer formatting on the return
    return loads(res)


def partial_runtime_comparison(
    baseline: dict[str, float],
    latest: dict[str, float],
    metrics: List[str],
    threshold: Optional[float] = 0.10,
) -> dict[str, str]:
    """
    interface into rust module
    serves to nicely formats the return as dicts are ordered and hashmaps are not
    Args:
        baseline: dict -> the result from calling perform_analysis on the baseline data
        latest: dict -> the current data for comparison from calling perform_analysis
        metrics: List[str] -> the list of metrics we want to evaluate on
        threshold: Optionl[float]=None -> the comparison threshold, defaults to 0.10 in rust mod
    Returns:
        dict
    """
    res: str = data_bias_partial_check(
        baseline=baseline, latest=latest, metrics=metrics, threshold=threshold
    )
    # for nicer formatting on the return
    return loads(res)



from pydantic import BaseModel, ConfigDict
from enum import Enum
from typing import Union


class ModelType(str, Enum):
    LinearRegression = "LinearRegression"
    LogisticRegression = "LogisticRegression"
    BinaryClassification = "BinaryClassification"


class ModelBiasBaseline(BaseModel):
    """data model for consistently formatted returns to users"""

    model_config = ConfigDict(extra="forbid", strict=True)
    DifferenceInPositivePredictedLabels: float
    DisparateImpact: float
    AccuracyDifference: float
    RecallDifference: float
    DifferenceInConditionalAcceptance: float
    DifferenceInAcceptanceRate: float
    SpecialityDifference: float
    DifferenceInConditionalRejection: float
    DifferenceInRejectionRate: float
    TreatmentEquity: float
    ConditionalDemographicDesparityPredictedLabels: float
    GeneralizedEntropy: float


class DataBiasBaseline(BaseModel):
    """data model for consistently formatted returns to users"""

    model_config = ConfigDict(extra="forbid", strict=True)
    ClassImbalance: float
    DifferenceInProportionOfLabels: float
    KlDivergence: float
    JsDivergence: float
    LpNorm: float
    TotalVarationDistance: float
    KolmorogvSmirnov: float


# some models for conistent formatting on metric objects
class LinearRegressionReport(BaseModel):
    RootMeanSquaredError: float
    MeanSquaredError: float
    MeanAbsoluteError: float
    RSquared: float
    MaxError: float
    MeanSquaredLogError: float
    RootMeanSquaredLogError: float
    MeanAbsolutePercentageError: float


class LogisticRegressionReport(BaseModel):
    BalancedAccuracy: float
    PrecisionPositive: float
    PrecisionNegative: float
    RecallPositive: float
    RecallNegative: float
    Accuracy: float
    F1Score: float
    LogLoss: float


class BinaryClassificationReport(BaseModel):
    BalancedAccuracy: float
    PrecisionPositive: float
    PrecisionNegative: float
    RecallPositive: float
    RecallNegative: float
    Accuracy: float
    F1Score: float


class ModelPerformance(BaseModel):
    model_config = ConfigDict(extra="forbid", strict=True, use_enum_values=True)
    modelType: ModelType
    performanceData: Union[
        LinearRegressionReport, LogisticRegressionReport, BinaryClassificationReport
    ]
